import json


# todo@rain
# 1. pre-calculate resource spec for strategy patch merge
# 2. calcuate for resource completion

class K8sSpec:
    def __init__(self, spec_file):
        with open(spec_file) as f:
            self.content = json.load(f)
            self.definitions = self.content['definitions']

    def _get_one_definition(self, resourceType):
        return self.definitions.get(resourceType)

    def _get_one_definition_properties(self, resourceType):
        return self._get_one_definition(resourceType).get('properties')

    def _get_one_definition_type(self, resourceType):
        return self._get_one_definition(resourceType).get('type')

    def gen_definition_tree(self, root_node):
        properties = self._get_one_definition_properties(root_node)
        results = {}
        if properties:
            for one_property_name, one_property_value in properties.items():
                if '$ref' in one_property_value:
                    results[one_property_name] = \
                        {**self.gen_definition_tree(one_property_value['$ref'].replace('#/definitions/', '')),
                         '_definition': one_property_value
                         }
                elif one_property_value.get('type') == 'array':
                    items =  one_property_value['items']
                    if '$ref' in  items:
                        results[one_property_name] = {
                            **self.gen_definition_tree(items['$ref'].replace('#/definitions/', '')),
                            '_definition' : one_property_value
                            }
                    else:
                        results[one_property_name] = {
                            '_definition' : one_property_value
                        }
                else:
                    results[one_property_name] = {
                        '_definition': one_property_value
                    }
        else:
            return {
                '_definition': self._get_one_definition(root_node)
            }
        return results

    def gen_sample(self, root_node):
        properties = self._get_one_definition_properties(root_node)
        results = {}
        if properties:
            for one_property_name, one_property_value in properties.items():
                if '$ref' in one_property_value:
                    results[one_property_name] = self.gen_sample(one_property_value['$ref'].replace('#/definitions/', ''))
                elif one_property_value.get('type') == 'array':
                    items =  one_property_value['items']
                    if '$ref' in  items:
                        results[one_property_name] = [self.gen_sample(items['$ref'].replace('#/definitions/', ''))]

                    else:
                        results[one_property_name] = ['']
                else:
                    results[one_property_name] = ''
        else:
            return ''
        return results


def pprint(obj):
    print(json.dumps(obj, indent=2))


if __name__ == '__main__':
    spec = K8sSpec('/Users/luganlin/git/mf/omc/omc/assets/k8s/swagger.json')

    # pprint(spec._get_one_definition_properties("io.k8s.api.apps.v1.Deployment"))
    # pprint(spec.gen_definition_tree("io.k8s.api.apps.v1.Deployment"))
    resource = 'deployment'
    with open('/Users/luganlin/git/mf/omc/omc/resources/kube/%s/_%s_completion.json' % (resource, resource), 'w') as f:
        json.dump(spec.gen_sample("io.k8s.api.apps.v1.%s" % resource.capitalize()), f, indent=2)
    # pprint(spec._get_one_definition_type("io.k8s.api.apps.v1.Deployment"))



import json

from omc_kube.core import StrategicMergePatch
from omc_kube.kube.kube_resource import KubeResource


class Deployment(KubeResource):
    def pause(self):
        resource_content = self._get_resource()
        namespace = self._get_namespace()
        config_key = 'spec.template.spec.containers[0]'
        config_value = json.loads('{"command": ["/bin/sh"], "args": ["-c", "while true; do echo hello; sleep 10;done"]}')
        patch_func = getattr(self.client, 'patch_namespaced_' + self._get_kube_api_resource_type())
        # patch_object = utils.build_object(config_key, config_value)
        patch_object = StrategicMergePatch.get_instance().gen_strategic_merge_patch(resource_content, config_key.split('.'),
                                                                                    config_value, 'set', [])
        new_result = patch_func(self._get_one_resource_value(), namespace, patch_object)
        print(new_result)


import argparse

from omc.core.decorator import filecache

from omc.common import CmdTaskMixin

from omc.core import Resource


class Container(Resource, CmdTaskMixin):
    def __init__(self, context={}, type='web'):
        super().__init__(context, type)
        self.client = self.context['common']['client']

    def _get_kube_resource_type(self):
        return self._get_resource_name()

    def _read_namespaced_resource(self, name, namespace, **kwargs):
        read_func = getattr(self.client, 'read_namespaced_' + self._get_kube_resource_type())
        return read_func(name, namespace, **kwargs)

    def _list_resource_for_all_namespaces(self):
        list_func = getattr(self.client, 'list_%s_for_all_namespaces' % self._get_kube_resource_type())
        return list_func()

    @filecache(duration=60 * 60, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(True))

        if not self._have_resource_value():
            pod_name = self._get_one_resource_value('pod')
            namespace = self.client.get_namespace('pod', pod_name)
            result = self.client.read_namespaced_pod(pod_name, namespace)
            # for one_container in result.spec.containers:
            results.extend(self._get_completion([(one.name,one.image) for one in result.spec.containers], False))

        return '\n'.join(results)

    def list(self):
        ret = self._list_resource_for_all_namespaces()
        print(ret)

    def describe(self):
        pass

    def exec(self):
        container_name = self._get_one_resource_value()
        pod_name = self._get_one_resource_value("pod")
        namespace = self.client.get_namespace('pod', pod_name)

        self.client.exec('pod', pod_name, namespace, " ".join(self._get_action_params()), container_name)

    def download(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--local', help='local dir')
        parser.add_argument('--remote', help='remote dir')
        args = parser.parse_args(self._get_action_params())

        pod_name = self._get_one_resource_value('pod')
        container_name = self._get_one_resource_value()
        namespace = self.client.get_namespace('pod', pod_name)
        self.client.download(pod_name, namespace, args.local, args.remote, container_name)

    def upload(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--local', help='local dir')
        parser.add_argument('--remote', help='remote dir')
        args = parser.parse_args(self._get_action_params())

        container_name = self._get_one_resource_value()
        pod_name = self._get_one_resource_value('pod')
        namespace = self.client.get_namespace('pod', pod_name)
        self.client.upload(pod_name, namespace, args.local, args.remote, container_name)



from omc.core.decorator import filecache

from omc.common import CmdTaskMixin

from omc.core import Resource


class Config(Resource, CmdTaskMixin):
    def __init__(self, context={}, type='web'):
        super().__init__(context, type)
        self.client = self.context['common']['client']

    def _get_kube_resource_type(self):
        return self._get_resource_name()

    def _read_namespaced_resource(self, name, namespace, **kwargs):
        read_func = getattr(self.client, 'read_namespaced_' + self._get_kube_resource_type())
        return read_func(name, namespace, **kwargs)

    def _list_resource_for_all_namespaces(self):
        list_func = getattr(self.client, 'list_%s_for_all_namespaces' % self._get_kube_resource_type())
        return list_func()

    @filecache(duration=60 * 60, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(True))

        if not self._have_resource_value():
            pod_name = self._get_one_resource_value('pod')
            namespace = self.client.get_namespace('pod', pod_name)
            result = self.client.read_namespaced_pod(pod_name, namespace)
            # for one_container in result.spec.containers:
            results.extend(self._get_completion([(one.name,one.image) for one in result.spec.containers], False))

        return '\n'.join(results)

    def get(self):
        pass



import os
import shutil

from omc.config import settings

from omc.common import CmdTaskMixin
from omc.core.resource import Resource
from omc.utils import file_utils


class Config(Resource, CmdTaskMixin):
    def get(self):
        k8s_resource = self._get_one_resource_value('kube')
        kube_config_dir = settings.OMC_KUBE_CONFIG_DIR

        the_kube_config_dir = os.path.join(kube_config_dir, k8s_resource)

        if os.path.exists(the_kube_config_dir):
            shutil.rmtree(the_kube_config_dir)

        file_utils.make_directory(the_kube_config_dir)

        # download config file
        print("downloading config file to %s" % the_kube_config_dir)

        config_download_cmd = 'scp %s:/root/.kube/config %s' % (k8s_resource, the_kube_config_dir)
        self.run_cmd(config_download_cmd)

        # download ssl files
        print("downloading ssl certificate to %s" % the_kube_config_dir)
        certificate_download_cmd = 'scp -r %s:/opt/kubernetes/ssl/ %s' % (k8s_resource, the_kube_config_dir)
        self.run_cmd(certificate_download_cmd)

        kube_config_file = os.path.join(the_kube_config_dir, 'config')
        file_utils.inplace_replace(kube_config_file, '/opt/kubernetes/ssl', os.path.join(the_kube_config_dir, 'ssl'))



import json
import os
from datetime import datetime

import pkg_resources

from omc.core.decorator import filecache

from omc.config import settings
from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

from omc.common import CmdTaskMixin
from omc.core.resource import Resource
from omc.utils import utils
from omc.utils.file_utils import make_directory
from omc.utils.utils import get_obj_value, get_all_dict_Keys, set_obj_value, delete_obj_key

from omc_kube.core import StrategicMergePatch


def dateconverter(o):
    if isinstance(o, datetime):
        return o.__str__()


class KubeNodeResource(Resource, CmdTaskMixin):
    def __init__(self, context={}, type='web'):
        super().__init__(context, type)
        self.client = self.context['common']['client']

    def _get_kube_resource_type(self):
        return self._get_parent_resource_name()

    def _get_kube_api_resource_type(self):
        return self._get_parent_resource_name()

    def _read_namespaced_resource(self, name, namespace, **kwargs):
        read_func = getattr(self.client, 'read_namespaced_' + self._get_kube_api_resource_type())
        return read_func(name, namespace, **kwargs)

    def _list_resource_for_all_namespaces(self):
        list_func = getattr(self.client, 'list_%s_for_all_namespaces' % self._get_kube_api_resource_type())
        return list_func()

    @filecache(duration=60 * 60, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(False))

        if not self._have_resource_value():
            # parent_resource = self._get_one_resource_value(self._get_kube_resource_type())
            # namespace = self.client.get_namespace(self._get_kube_api_resource_type(), parent_resource)

            kube_resource_type = self._get_kube_resource_type()
            completion_file_name = pkg_resources.resource_filename('omc_kube.%(kube_resource_type)s' % locals(), '_%(kube_resource_type)s_completion.json' % locals())
            prompts = []
            # get_all_dict_Keys(result.to_dict(), prompts)
            with open(completion_file_name) as f:
                result = json.load(f)
                get_all_dict_Keys(result, prompts)
                results.extend(self._get_completion(prompts))

        return "\n".join(results)

    @staticmethod
    def _build_field_selector(selectors):
        return ','.join(['%s=%s' % (k, v) for (k, v) in selectors.items()])

    def get(self):
        'get resource by configuration key'
        parent_resource = self._get_one_resource_value(self._get_kube_resource_type())
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), parent_resource)
        result = self._read_namespaced_resource(parent_resource, namespace)

        # config resource is the item
        resource = self._get_one_resource_value()

        if not resource:
            print(result)
        else:
            print(get_obj_value(result, resource))

    def set(self):
        'update restore by configuration key'
        # https://github.com/kubernetes/kubernetes/blob/cea1d4e20b4a7886d8ff65f34c6d4f95efcb4742/staging/src/k8s.io/apimachinery/pkg/util/strategicpatch/patch.go
        # implement Strategic merge patch myself
        # https://stackoverflow.com/questions/36307950/kubernetes-api-call-equivalent-to-kubectl-apply
        if 'completion' in self._get_params():
            resource = self._get_one_resource_value()
            namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
            result = self._read_namespaced_resource(resource, namespace)
            prompts = []
            get_all_dict_Keys(result.to_dict(), prompts)
            self._print_completion(prompts)
            return

        config_key = self._get_one_resource_value()
        parent_resource = self._get_one_resource_value(self._get_kube_resource_type())
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), parent_resource)
        result = self._read_namespaced_resource(parent_resource, namespace)
        params = self._get_action_params()
        config_value = params[0]
        orig_value = get_obj_value(result, config_key)
        # convert type
        config_value = type(orig_value)(config_value)
        set_obj_value(result, config_key, config_value)

        # todo: use apply instead once apply provided
        patch_func = getattr(self.client, 'patch_namespaced_' + self._get_kube_api_resource_type())
        # patch_object = utils.build_object(config_key, config_value)
        patch_object = StrategicMergePatch.get_instance().gen_strategic_merge_patch(result, config_key.split('.'),
                                                                                    config_value, 'set', [])
        new_result = patch_func(parent_resource, namespace, patch_object)
        print(new_result)

    def delete(self):
        'delete node by configuration key'
        # todo@rain: to support delete entired resource and completion cache
        if 'completion' in self._get_params():
            resource = self._get_one_resource_value()
            namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
            result = self._read_namespaced_resource(resource, namespace)
            prompts = []
            get_all_dict_Keys(result.to_dict(), prompts)
            self._print_completion(prompts)
            return

        config_key = self._get_one_resource_value()
        parent_resource = self._get_one_resource_value(self._get_kube_resource_type())
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), parent_resource)
        the_result = self._read_namespaced_resource(parent_resource, namespace, _preload_content=False)
        result = json.loads(the_result.data.decode('UTF-8'))
        params = self._get_action_params()
        config_value = params[0] if len(params) > 0 else None
        # orig_value = get_obj_value(result, config_key)
        # # convert type
        # config_value = type(orig_value)(config_value)
        # set_obj_value(result, config_key, config_value)

        # todo: use apply instead once apply provided
        patch_func = getattr(self.client, 'patch_namespaced_' + self._get_kube_api_resource_type())
        # patch_object = utils.build_object(config_key, config_value)
        patch_object = StrategicMergePatch.get_instance().gen_strategic_merge_patch(result, config_key.split('.'),
                                                                                    config_value, 'delete', [])
        new_result = patch_func(parent_resource, namespace, patch_object)
        print(new_result)

    def _edit(self):
        'Edit a resource from the default editor.'
        resource = self._get_one_resource_value(self._get_kube_resource_type())
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)

        self.client.edit(self._get_kube_resource_type(), resource, namespace)

    def _save(self):
        'save configuration in file cache to be restored'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)
        kube_instance = self._get_one_resource_value("kube")
        if not kube_instance:
            kube_instance = 'local'
        cache_folder = os.path.join(settings.OMC_KUBE_CACHE_DIR, kube_instance, namespace,
                                    self._get_kube_resource_type())

        result = self._read_namespaced_resource(resource_name, namespace, _preload_content=False)
        stream = StringIO()
        the_result = json.loads(result.data.decode('UTF-8'))
        delete_obj_key(the_result, 'metadata.creationTimestamp')
        delete_obj_key(the_result, 'metadata.resourceVersion')
        yaml = YAML()
        yaml.dump(the_result, stream)
        content = stream.getvalue()

        make_directory(cache_folder)
        with open(os.path.join(cache_folder, resource_name + '.yaml'), 'w') as f:
            f.write(content)

    def _restore(self):
        'restore configuration saved in file cache'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)
        kube_instance = self._get_one_resource_value("kube")
        if not kube_instance:
            kube_instance = 'local'
        cache_folder = os.path.join(settings.OMC_KUBE_CACHE_DIR, kube_instance, namespace,
                                    self._get_kube_resource_type())
        make_directory(cache_folder)

        config_file = os.path.join(cache_folder, resource_name + '.yaml')
        if os.path.exists(config_file):
            self.client.apply(config_file)
        else:
            raise Exception("no config file found")

if __name__ ==  '__main__':
    resource = pkg_resources.resource_filename('omc_kube.deployment', '_deployment_completion.json')
    print(resource)


import os

from omc.common import CmdTaskMixin
from omc.config import settings
from omc.core.decorator import filecache
from omc.core.resource import Resource

from omc_kube.core import KubernetesClient


class Kube(Resource, CmdTaskMixin):
    def _description(self):
        return 'The Kubernetes command-line tool'

    @filecache(duration=-1, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(False))
        if not self._have_resource_value():
            if os.path.exists(settings.OMC_KUBE_CONFIG_DIR):
                resources = os.listdir(settings.OMC_KUBE_CONFIG_DIR)
                results.append("\n".join(self._get_completion(resources, True)))

        return "\n".join(results)

    def _before_sub_resource(self):
        try:
            if self._have_resource_value():
                resource_value = self._get_one_resource_value()
                client = KubernetesClient(os.path.join(settings.OMC_KUBE_CONFIG_DIR, resource_value, 'config'))
            else:
                client = KubernetesClient()

            self.context['common'] = {'client': client}
        except Exception as e:
            # some action no need to create load config, get config action e.g.
            raise Exception('invalid kubenetes config')


if __name__ == '__main__':
    client = KubernetesClient()
    ret = client.list_service_for_all_namespaces(watch=False)

    # print(client.read_namespaced_pod("postgres-svc-5685d4bc7-l6j4m", 'default'))
    # print(client.read_namespaced_pod_template("postgres-svc-5685d4bc7-l6j4m", 'default'))
    # print(client.read_namspaced_event("postgres-svc-5685d4bc7-l6j4m", 'default'))
    print(ret)
    # for i in ret.items:
    #     print("%s\t%s\t%s" % (i.status.pod_ip, i.metadata.namespace, i.metadata.name))



from omc.common import CmdTaskMixin
from omc_kube.kube.kube_node_resource import KubeNodeResource


class Node(KubeNodeResource, CmdTaskMixin):
    def _get_kube_api_resource_type(self):
        return 'config_map'




import os
import shutil

from omc.common import CmdTaskMixin
from omc.utils import file_utils
from omc.utils.object_utils import ObjectUtils


class KubeConfigService(CmdTaskMixin):
    def sync(self, k8s_resource, kube_config_dir):

        the_kube_config_dir = os.path.join(kube_config_dir, k8s_resource)

        if os.path.exists(the_kube_config_dir):
            shutil.rmtree(the_kube_config_dir)

        file_utils.make_directory(the_kube_config_dir)

        # download config file
        print("downloading config file to %s" % the_kube_config_dir)

        config_download_cmd = 'scp %s:/root/.kube/config %s' % (k8s_resource, the_kube_config_dir)
        self.run_cmd(config_download_cmd)

        ssl_dir = self._get_cert_dir(os.path.join(the_kube_config_dir, 'config'))

        # download ssl files
        print("downloading ssl certificate to %s" % the_kube_config_dir)
        certificate_download_cmd = 'scp -r %s:%s %s' % (k8s_resource,ssl_dir, the_kube_config_dir)
        self.run_cmd(certificate_download_cmd)

        kube_config_file = os.path.join(the_kube_config_dir, 'config')
        file_utils.inplace_replace(kube_config_file, ssl_dir, os.path.join(the_kube_config_dir, 'ssl'))

    def _get_cert_dir(self, config_file):
        import yaml
        with open(config_file) as f:
            result = yaml.load(f, yaml.FullLoader)
            path = (ObjectUtils.get_node(result, 'clusters[0].cluster.certificate-authority'))
            return os.path.dirname(path)

if __name__ == '__main__':
    import yaml
    with open('/Users/luganlin/.omc/config/kube/cd219/config') as f:
        result = yaml.load(f, yaml.FullLoader)
        path = (ObjectUtils.get_node(result, 'clusters[0].cluster.certificate-authority'))
        print(os.path.dirname(path))
        print(result)


import argparse

from omc_kube.kube.kube_resource import KubeResource


class Pod(KubeResource):
    pass

    def download(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--local', help='local dir')
        parser.add_argument('--remote', help='remote dir')
        args = parser.parse_args(self._get_action_params())

        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)
        self.client.download(resource_name, namespace, args.local, args.remote)

    def upload(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--local', help='local dir')
        parser.add_argument('--remote', help='remote dir')
        args = parser.parse_args(self._get_action_params())

        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)
        self.client.upload(resource_name, namespace, args.local, args.remote)



import json
import os
from datetime import datetime
import logging
from omc.core.decorator import filecache

from omc.config import settings
from ruamel.yaml import YAML
from ruamel.yaml.compat import StringIO

from omc.common import CmdTaskMixin
from omc.core.resource import Resource
from omc.utils.file_utils import make_directory
from omc.utils.utils import get_obj_value, get_all_dict_Keys, set_obj_value, delete_obj_key


def dateconverter(o):
    if isinstance(o, datetime):
        return o.__str__()


class KubeResource(Resource, CmdTaskMixin):
    def __init__(self, context={}, type='web'):
        super().__init__(context, type)
        self.client = self.context['common']['client']

    def _get_kube_resource_type(self):
        return self._get_resource_name()

    def _get_kube_api_resource_type(self):
        # specified for python api,
        # same as resource type by default
        # but some resource type is slight different configmap -> config_map
        return self._get_resource_name()

    def _read_namespaced_resource(self, name, namespace, **kwargs):
        read_func = getattr(self.client, 'read_namespaced_' + self._get_kube_api_resource_type())
        return read_func(name, namespace, **kwargs)

    def _list_resource_for_all_namespaces(self):
        list_func = getattr(self.client, 'list_%s_for_all_namespaces' % self._get_kube_api_resource_type())
        return list_func()

    def _get_resource(self):
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
        return self._read_namespaced_resource(resource, namespace)

    def _get_namespace(self):
        resource = self._get_one_resource_value()
        return self.client.get_namespace(self._get_kube_api_resource_type(), resource)

    @filecache(duration=60 * 60, file=Resource._get_cache_file_name)
    def _completion(self, short_mode=True):
        results = []
        results.append(super()._completion(False))

        if not self._have_resource_value():
            ret = self._list_resource_for_all_namespaces()
            results.extend(
                self._get_completion([get_obj_value(one, 'metadata.name') for one in ret.get('items')], True))

        return "\n".join(results)

    def list(self):
        'display one or more resources'
        resource_name = self._get_one_resource_value()
        namespace = 'all' if not resource_name else self.client.get_namespace(self._get_kube_api_resource_type(),
                                                                              resource_name)

        # ret = self._list_resource_for_all_namespaces()
        # print(ret)
        result = self.client.get(self._get_kube_resource_type(), resource_name, namespace)
        print(result)

    def yaml(self):
        'get configuration in yaml format'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        stream = StringIO()
        yaml = YAML()
        yaml.dump(result, stream)
        print(stream.getvalue())

    def json(self):
        'get configuration by json format'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        print(json.dumps(result, default=dateconverter, indent=4))

    @staticmethod
    def _build_field_selector(selectors):
        return ','.join(['%s=%s' % (k, v) for (k, v) in selectors.items()])

    def namespace(self):
        'get resource namespace'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
        print(namespace)

    def event(self):
        'show event on the resource'
        # https://kubernetes.docker.internal:6443/api/v1/namespaces/default/events?fieldSelector=
        # involvedObject.uid=4bb31f4d-99f1-4acc-a024-8e2484573733,
        # involvedObject.name=itom-xruntime-rabbitmq-6464654786-vnjxz,
        # involvedObject.namespace=default

        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)

        uid = get_obj_value(result, 'metadata.uid')
        name = get_obj_value(result, 'metadata.name')

        the_selector = {
            "involvedObject.uid": uid,
            "involvedObject.namespace": namespace,
            "involvedObject.name": name,
        }

        print(self.client.list_namespaced_event(namespace, field_selector=self._build_field_selector(the_selector)))

    def _get_config_key_cache_file_name(self):
        main_path = [one for one in self.context['all'][1:] if not one.startswith('-')]
        cache_file = os.path.join(settings.OMC_COMPLETION_CACHE_DIR, *main_path)
        return cache_file

    @filecache(duration=60 * 5, file=_get_config_key_cache_file_name)
    def _get_config_key_completion(self):
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
        result = self._read_namespaced_resource(resource, namespace)
        prompts = []
        get_all_dict_Keys(result.to_dict(), prompts)
        return '\n'.join(self._get_completion(prompts))

    # def get(self):
    #     'get resource by configuration key'
    #     if 'completion' in self._get_params():
    #         completion = self._get_config_key_completion()
    #         print(completion)
    #         return
    #
    #     resource = self._get_one_resource_value()
    #     namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
    #     result = self._read_namespaced_resource(resource, namespace)
    #     params = self._get_action_params()
    #
    #     the_params = " ".join(params)
    #
    #     if not the_params.strip():
    #         print(result)
    #     else:
    #         print(get_obj_value(result, the_params))
    #
    # def set(self):
    #     'update restore by configuration key'
    #     if 'completion' in self._get_params():
    #         resource = self._get_one_resource_value()
    #         namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
    #         result = self._read_namespaced_resource(resource, namespace)
    #         prompts = []
    #         get_all_dict_Keys(result.to_dict(), prompts)
    #         self._print_completion(prompts)
    #         return
    #
    #     resource = self._get_one_resource_value()
    #     namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
    #     result = self._read_namespaced_resource(resource, namespace)
    #     params = self._get_action_params()
    #     config_key = params[0]
    #     config_value = params[1]
    #     orig_value = get_obj_value(result, config_key)
    #     # convert type
    #     config_value = type(orig_value)(config_value)
    #     set_obj_value(result, config_key, config_value)
    #
    #     # todo: use apply instead once apply provided
    #     new_result = self.client.replace_namespaced_deployment(resource, namespace, result)
    #     print(get_obj_value(new_result, config_key))
    #
    # def delete(self):
    #     'delete node by configuration key'
    #     # todo@rain: to support delete entired resource and completion cache
    #     if 'completion' in self._get_params():
    #         resource = self._get_one_resource_value()
    #         namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
    #         result = self._read_namespaced_resource(resource, namespace)
    #         prompts = []
    #         get_all_dict_Keys(result.to_dict(), prompts)
    #         self._print_completion(prompts)
    #         return
    #
    #     resource = self._get_one_resource_value()
    #     namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)
    #     result = self._read_namespaced_resource(resource, namespace)
    #     params = self._get_action_params()
    #     config_key = params[0]
    #     # convert type
    #     delete_obj_key(result, config_key)
    #
    #     # todo: use apply instead once apply provided
    #     new_result = self.client.replace_namespaced_deployment(resource, namespace, result)

    def edit(self):
        'Edit a resource from the default editor.'
        resource = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource)

        self.client.edit(self._get_kube_resource_type(), resource, namespace)

    def save(self):
        'save configuration in file cache to be restored'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)
        kube_instance = self._get_one_resource_value("kube")
        if not kube_instance:
            kube_instance = 'local'
        cache_folder = os.path.join(settings.OMC_KUBE_CACHE_DIR, kube_instance, namespace,
                                    self._get_kube_resource_type())

        result = self._read_namespaced_resource(resource_name, namespace, _preload_content=False)
        stream = StringIO()
        the_result = json.loads(result.data.decode('UTF-8'))
        delete_obj_key(the_result, 'metadata.creationTimestamp')
        delete_obj_key(the_result, 'metadata.resourceVersion')
        delete_obj_key(the_result, 'metadata.annotations')
        delete_obj_key(the_result, 'metadata.managedFields')
        delete_obj_key(the_result, 'metadata.uid')
        delete_obj_key(the_result, 'metadata.selfLink')
        delete_obj_key(the_result, 'metadata.generation')
        delete_obj_key(the_result, 'status')
        yaml = YAML()
        yaml.dump(the_result, stream)
        content = stream.getvalue()

        make_directory(cache_folder)
        config_file = os.path.join(cache_folder, resource_name + '.yaml')
        logging.info('save %s %s to %s' % (self._get_kube_resource_type(), resource_name, config_file))
        with open(config_file, 'w') as f:
            f.write(content)

    def restore(self):
        'restore configuration saved in file cache'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)
        kube_instance = self._get_one_resource_value("kube")
        if not kube_instance:
            kube_instance = 'local'
        cache_folder = os.path.join(settings.OMC_KUBE_CACHE_DIR, kube_instance, namespace,
                                    self._get_kube_resource_type())
        make_directory(cache_folder)

        config_file = os.path.join(cache_folder, resource_name + '.yaml')

        logging.info('restore %s %s from %s' % (self._get_kube_resource_type(), resource_name, config_file))
        if os.path.exists(config_file):
            self.client.apply(config_file)
        else:
            raise Exception("config file %s not found" % config_file)

    def exec(self):
        'Execute a command in a container'
        resource_name = self._get_one_resource_value()
        namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)

        self.client.exec(self._get_kube_resource_type(), resource_name, namespace, " ".join(self._get_action_params()))

    def describe(self):
        'Show details of a specific resource or group of resources'
        resource_name = self._get_one_resource_value()
        namespace = 'all'
        if resource_name:
            namespace = self.client.get_namespace(self._get_kube_api_resource_type(), resource_name)

        print(self.client.describe(self._get_kube_resource_type(), resource_name, namespace))

    def relations(self):
        pass



import json
import re

from kubernetes import client, config
from omc.common import CmdTaskMixin
from functools import wraps


def get_obj_value(obj, key):
    # e.g. get pod.data.ips[0]
    if not key:
        return None

    first_attr, others, delimiter = extract_first_attr(key)

    if first_attr:
        if isinstance(obj, dict):
            first_value = obj.get(first_attr)
        elif isinstance(obj, list):
            first_value = obj[int(first_attr)]
        else:
            first_value = getattr(obj, first_attr)

        if not others:
            return first_value

        else:
            return get_obj_value(first_value, others)


def extract_first_attr(key):
    key = key.strip('[].')
    for index in range(0, len(key)):
        if key[index] in '[].':
            return key[:index], key[index:].strip('[].'), key[index]

    return key, None, None


class KubernetesClient(CmdTaskMixin):
    def __init__(self, config_file=None):
        if not config_file:
            config.load_kube_config()
        else:
            config.load_kube_config(config_file)

        self.config_file = config_file

        corev1api = client.CoreV1Api()
        appsv1api = client.AppsV1Api()
        extensionsv1beta1api = client.ExtensionsV1beta1Api()
        self.client_instances = [corev1api, appsv1api, extensionsv1beta1api]

    def __getattr__(self, item: str):
        for one_instance in self.client_instances:
            if hasattr(one_instance, item):
                f = getattr(one_instance, item)

                @wraps(f)
                def wrapper(*args, **kwds):
                    if kwds is not None and '_preload_content' in kwds:
                        return f(*args, **kwds)
                    else:
                        result = f(*args, **kwds, _preload_content=False)
                        the_result = json.loads(result.data.decode('UTF-8'))
                        return the_result

                return wrapper

    def get_namespace(self, resource_type: str, resource_name: str):
        list_namespaces = getattr(self, "list_%s_for_all_namespaces" % resource_type)
        all_namespaces = list_namespaces()
        for one in all_namespaces.get('items'):
            if utils.get_obj_value(one, 'metadata.name') == resource_name:
                return utils.get_obj_value(one, 'metadata.namespace')

    ##################################
    ###### kubectl impl ##############
    ##################################

    def apply(self, file):
        # todo: since apply not supported, need to impl

        # option1: using kubectl instead
        # option2: upgrade k8s version to support server-side apply
        # option3: implemented in client side myself
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''

        cmd = "kubectl %(config)s apply -f %(file)s" % locals()
        result = self.run_cmd(cmd)

    def edit(self, resource_type: str, resource_name: str, namespace: str):
        resource_type = self.reconcile_resource_type(resource_type)
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        self.run_cmd("kubectl %(config)s edit %(resource_type)s %(resource_name)s --namespace %(namespace)s" % locals())

    def portforward(self):
        pass

    def download(self, resource_name, namespace, local_dir, remote_dir, container=None):
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        container_options = '' if not container else '-c ' + container
        cmd = "kubectl %(config)s cp %(container_options)s %(namespace)s/%(resource_name)s:%(remote_dir)s %(local_dir)s" % locals()
        self.run_cmd(cmd)

    def upload(self, resource_name, namespace, local_dir, remote_dir, container=None):

        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        container_options = '' if not container else '-c ' + container
        cmd = "kubectl %(config)s cp %(container_options)s %(local_dir)s %(namespace)s/%(resource_name)s:%(remote_dir)s" % locals()
        self.run_cmd(cmd)

    def exec(self, resource_type, resource_name, namespace, command, container=None, stdin=True):
        resource_type = self.reconcile_resource_type(resource_type)
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        interactive_option = '-it' if stdin else ''
        container_options = '' if not container else '-c ' + container
        cmd = 'kubectl %(config)s exec %(interactive_option)s %(resource_type)s/%(resource_name)s %(container_options)s --namespace %(namespace)s -- %(command)s' % locals()
        self.run_cmd(cmd)

    def get(self, resource_type, resource_name='', namespace='all', output='yaml'):
        resource_type = self.reconcile_resource_type(resource_type)
        resource_name = '' if not resource_name else resource_name
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        namespace_options = ''
        if namespace == 'all':
            namespace_options = '-A'
        else:
            namespace_options = '--namespace %s' % namespace

        output_options = '' if not output else '-o ' + output

        cmd = 'kubectl %(config)s get %(resource_type)s %(resource_name)s %(namespace_options)s -o wide' % locals()
        result = self.run_cmd(cmd, capture_output=True, verbose=False)
        return result.stdout.decode('UTF-8')

    def describe(self, resource_type, resource_name=None, namespace='all'):
        resource_type = self.reconcile_resource_type(resource_type)
        config = ' --kubeconfig %s ' % self.config_file if self.config_file else ''
        namespace_options = ''
        if namespace == 'all':
            namespace_options = '-A'
        else:
            namespace_options = '--namespace %s' % namespace

        resource_name = '' if not resource_name else resource_name
        cmd = 'kubectl %(config)s describe %(resource_type)s %(resource_name)s %(namespace_options)s' % locals()
        result = self.run_cmd(cmd, capture_output=True, verbose=False)
        return result.stdout.decode("UTF-8")

    def reconcile_resource_type(self, resource_type):
        resource_type_mapping = {
            'config_map': 'configmap'
        }

        if resource_type in resource_type_mapping:
            return resource_type_mapping[resource_type]
        else:
            return resource_type


# https://github.com/kubernetes/community/blob/master/contributors/devel/sig-api-machinery/strategic-merge-patch.md
# openapi spec https://raw.githubusercontent.com/kubernetes/kubernetes/master/api/openapi-spec/swagger.json
#

'''
{
    "spec": {
        "template": {
            "spec": {
                "$setElementOrder/containers": [
                    {
                        "name": "idm"
                    }
                ],
                "containers": [
                    {
                        "$setElementOrder/env": [
                            {
                                "name": "EXTERNAL_NAME"
                            },
                            {
                                "name": "EXTERNAL_NAME1"
                            },
                            {
                                "name": "IDM_MOUNT_ROOT"
                            },
                            {
                                "name": "JVM_MAX_SIZE"
                            },
                            {
                                "name": "CATALINA_OPTS"
                            },
                            {
                                "name": "HPSSO_INIT_STRING"
                            },
                            {
                                "name": "SAML2_ENABLE"
                            },
                            {
                                "name": "SAML_KEYSTORE_PASSWORD_KEY"
                            },
                            {
                                "name": "saml_keystore_password_key"
                            }
                        ],
                        "env": [
                            {
                                "name": "EXTERNAL_NAME1",
                                "value": "idm-svc1"
                            }
                        ],
                        "name": "idm"
                    }
                ]
            }
        }
    }
}
'''


class StrategicMergePatch:
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self.api_spec = {
            'spec.template.spec.containers': {
                "type": "array",
                "x-kubernetes-patch-merge-key": "name",
                "x-kubernetes-patch-strategy": "merge"
            },
            'spec.template.spec.containers.env': {
                "type": "array",
                "x-kubernetes-patch-merge-key": "name",
                "x-kubernetes-patch-strategy": "merge"
            },

        }

    def delete_obj_key(self, obj, key):
        if not key:
            return None

        first_attr, others, delimiter = self.extract_first_attr(key)

        if not others:
            # do set value

            if first_attr:
                if isinstance(obj, dict):
                    # obj.pop(key, None)
                    del obj[key]
                elif isinstance(obj, list):
                    del obj[int(first_attr)]
                else:
                    try:
                        delattr(obj, first_attr)
                    except:
                        setattr(obj, first_attr, None)

                return
        else:
            # do get value

            if first_attr:
                if isinstance(obj, dict):
                    first_value = obj.get(first_attr)
                elif isinstance(obj, list):
                    first_value = obj[int(first_attr)]
                else:
                    first_value = getattr(obj, first_attr)

                self.delete_obj_key(first_value, others)

    def set_obj_value(self, obj, key, value):
        # e.g. get pod.data.ips[0]
        if not key:
            return None

        first_attr, others, delimiter = self.extract_first_attr(key)

        if not others:
            # do set value

            if first_attr:
                if isinstance(obj, dict):
                    obj[first_attr] = value
                elif isinstance(obj, list):
                    obj[int(first_attr)] = value
                else:
                    setattr(obj, first_attr, value)

                return
        else:
            # do get value

            if first_attr:
                if isinstance(obj, dict):
                    first_value = obj.get(first_attr)
                elif isinstance(obj, list):
                    first_value = obj[int(first_attr)]
                else:
                    first_value = getattr(obj, first_attr)

                self.set_obj_value(first_value, others, value)

    def get_obj_value(self, obj, key):
        # e.g. get pod.data.ips[0]
        if not key:
            return None

        first_attr, others, delimiter = self.extract_first_attr(key)

        if first_attr:
            if isinstance(obj, dict):
                first_value = obj.get(first_attr)
            elif isinstance(obj, list):
                first_value = obj[int(first_attr)]
            else:
                first_value = getattr(obj, first_attr)

            if not others:
                return first_value

            else:
                return self.get_obj_value(first_value, others)

    def extract_first_attr(self, key):
        key = key.strip('[].')
        for index in range(0, len(key)):
            if key[index] in '[].':
                return key[:index], key[index:].strip('[].'), key[index]

        return key, None, None

    def get_all_dict_Keys(self, obj, paths=[]):
        if isinstance(obj, dict):
            for k, v in obj.items():
                paths.append(k)
                subpaths = []
                self.get_all_dict_Keys(v, subpaths)
                paths.extend([k + one if one.startswith('[') else k + '.' + one for one in subpaths])
        elif isinstance(obj, list):
            paths.append('[0]')
            subpaths = []
            self.get_all_dict_Keys(obj[0], subpaths)
            paths.extend(['[0].' + one for one in subpaths])

    def build_object(self, key, value, init_object={}):
        # todo: to support array
        first, others, delimiter = self.extract_first_attr(key)

        if others is None:
            # build_object('a', 'b')
            if isinstance(init_object, list):
                index = int(first)
                if index < len(init_object):
                    init_object[index] = value
                elif index == len(init_object):
                    init_object.append(value)
            else:
                init_object[key] = value
            return init_object
        elif delimiter == '.':
            # build_object('a.b', 'c')
            return {first: self.build_object(others, value, {})}
        else:
            if isinstance(init_object, list):
                index = int(first)
                if index < len(init_object):
                    init_object[int(first)] = self.build_object(others, value, {})
                    return init_object
                elif index == len(init_object):
                    init_object.append(self.build_object(others, value, {}))
                    return init_object
                else:
                    raise Exception('out of range')
            else:
                # build_object('a[0].b', 'c')
                init_object = {first: self.build_object(others, value, [])}
                return init_object

    def flatten_key(self, key):
        return re.sub("\[\d+\]", '', key)

    def parse_one_key(self, key):
        is_array = False
        index = -1
        key = key
        if '[' in key and ']' in key:
            is_array = True
            parsed_key, parsed_index = re.findall('(\w+)\[(\d*)\]', key)[0]
            key = parsed_key

            if parsed_index:
                index = int(parsed_index)
            else:
                index = -1

        return key, is_array, index
        # return {
        #     'key': key,
        #     'is_array': is_array,
        #     'index': index
        # }

    def gen_strategic_merge_patch(self, origin, key_array, value, action, path):
        '''
        'spec.template.spec.containers[0].env[1]',
         {'name': 'name1', 'value': 'value1'}
        '''
        keys = key_array.copy()
        one_key = keys.pop(0)
        current_path = path.copy()  # raw path, with array index
        current_path.append(one_key)

        the_key, is_array, index = self.parse_one_key(one_key)
        current_key_path = '.'.join(
            [*path, the_key])  # with old array index, without current index, for get object value
        current_flatten_path = self.flatten_key('.'.join([*path, the_key]))  # no array index, only for api spec usage

        if not is_array:
            if not keys:
                # no key any more
                if action == 'set':
                    return {the_key: value}
                elif action == 'delete':
                    return {the_key: None}
            else:
                return {the_key: self.gen_strategic_merge_patch(origin, keys, value, action, current_path)}
        else:

            the_value = value

            if current_flatten_path in self.api_spec:
                # for array, we have stratgic merge rule
                current_value = self.get_obj_value(origin, current_key_path)

                merge_key = self.api_spec[current_flatten_path]['x-kubernetes-patch-merge-key']
                patch_strategy = self.api_spec[current_flatten_path]['x-kubernetes-patch-strategy']
                results = {}

                '''
                 "$setElementOrder/containers": [
                    {
                        "name": "idm"
                    }
                ],
                '''
                if keys:
                    # has children
                    results['$setElementOrder/%s' % the_key] = [{merge_key: one.get(merge_key)} for one in
                                                                current_value]
                    results[the_key] = [
                        {
                            merge_key: current_value[index].get(merge_key),
                            **self.gen_strategic_merge_patch(origin, keys, the_value, action, current_path)
                        }
                    ]

                if not keys:
                    results['$setElementOrder/%s' % the_key] = [{merge_key: one.get(merge_key)} for one in
                                                                current_value]
                    if action == 'set':
                        one_item = []
                        if isinstance(value, dict):
                            value = {**value, merge_key: current_value[index].get(merge_key)}
                        one_item.insert(index, value)
                        results[the_key] = one_item
                    elif action == 'delete':
                        one_item = []
                        # one_item.insert(index, value)
                        one_item.append({"$patch": 'delete', **value})
                        results[the_key] = one_item
                return results

            else:
                # if no merge rule, merge normally
                current_value = self.get_obj_value(origin, current_key_path)
                current_value = current_value if current_value else []
                current_value.insert(index, the_value)


if __name__ == '__main__':
    # client = KubernetesClient("~/.omc/config/kube/nightly1/config")
    # the_namespace = client.get_namespace("service", "smarta-smart-ticket-svc")
    # print(the_namespace)
    #
    # print(client.list_config_map_for_all_namespaces(watch=False))
    # print(client.list_pod_for_all_namespaces(watch=False))
    # print(client.list_deployment_for_all_namespaces(watch=False))
    # print(client.list_service_for_all_namespaces(watch=False))
    # print(client.list_endpoints_for_all_namespaces(watch=False))
    smp = StrategicMergePatch()
    origin = None
    with open('/Users/luganlin/git/mf/omc/omc/assets/k8s/deployment_sample.json') as f:
        origin = json.load(f)
    # result = smp.gen_strategic_merge_patch(origin, 'spec.template.spec.containers[0].env[]'.split('.'),
    #                                        {'name': 'name1', 'value': 'value1'}, 'delete', [])

    # # print(json.dumps(result, indent=2))
    #
    result1 = smp.gen_strategic_merge_patch(origin, 'spec.template.spec.containers[0].livenessProbe'.split('.'),
                                            'value1', 'delete', [])

    print(json.dumps(result1, indent=2))
    value = json.loads('{"command": ["/bin/sh"], "args": ["-c", "while true; do echo hello; sleep 10;done"]}')
    result2 = smp.gen_strategic_merge_patch(origin, 'spec.template.spec.containers[0]'.split('.'),
                                            value, 'set', [])

    print(json.dumps(result2, indent=2))
    # print(smp.gen_strategic_merge_patch(origin, 'spec.template.spec.containers[0].env',{'name': 'name1', 'value': 'value1'}))

    # print(json.loads('{"command": ["/bin/sh"], "args": ["-c", "while true; do echo hello; sleep 10;done"]}'))



from setuptools import setup, find_packages

install_requires = [
    'omc>=0.2.3',
    'kubernetes>=11.0.0',
]

setup(
    name='omc-kube',
    version="0.0.4",
    description='kubenetes plugin for omc',
    license='MIT',
    author='Lu Ganlin',
    author_email='linewx1981@gmail.com',
    url='https://github.com/linewx/omc-kube',
    packages=find_packages(),
    package_data={'omc_kube.config': ['*.yaml'], 'omc_kube.lib': ['**', '**/*', '**/**/*']},
    install_requires=install_requires,
)
